"""
Consumer for RedisMQ
"""
from __future__ import annotations

import asyncio
import json
from functools import partial

from typing import TYPE_CHECKING, Any, Dict, TypedDict, Callable, Optional
from redis import Connection  # type: ignore[attr-defined]

from .debugging import debugging


class Client(TypedDict):
    redis: Connection


@debugging
class Consumer:  # pylint: disable=too-few-public-methods
    """
    Consumes messages from a stream
    """

    client: Client
    stream_name: str
    group_name: str
    consumer_name: str
    latest_id: bytes
    check_backlog: bool
    min_idle_time: int

    log_debug: Callable[..., None]

    def __init__(
        self,
        client: Client,
        stream_name: str,
        group_name: str,
        consumer_name: str,
        # scan_pending_on_start: bool = True,
        # claim_stale_messages: bool = True,
        min_idle_time: int = 60000,
    ) -> None:
        """
        default constructor
        """
        Consumer.log_debug(
            "__init__ %r %r %r %r", client, stream_name, group_name, consumer_name
        )

        self.client = client
        self.stream_name = stream_name
        self.group_name = group_name
        self.consumer_name = consumer_name
        # self.scan_pending_on_start = scan_pending_on_start
        # self.claim_stale_messages = claim_stale_messages
        self.min_idle_time = min_idle_time

        # by default just read new messages that haven't been delivered
        self.latest_id = b">"
        self.check_backlog = False
        self.xread_timeout = 10000  # milliseconds

    def read(self) -> PayloadFuture:
        """
        Read a message from the stream.
        """
        Consumer.log_debug("read(%s)", self.consumer_name)

        # create a future to hold the result
        read_result: PayloadFuture = asyncio.Future()

        # create a task to read the next message
        get_message_task = asyncio.create_task(self.get_message(read_result))
        Consumer.log_debug("    - get_message_task: %r", get_message_task)

        get_message_task.add_done_callback(self._get_message_task_callback)

        # add a callback to the read_result which will be called when the task
        # has received a message and wrapped it in a Payload, or if the
        # read_result gets canceled
        read_result.add_done_callback(partial(self._read_done, get_message_task))

        return read_result

    def _get_message_task_callback(self, *args):
        """
        Callback function for getting a message from the stream, used for
        debugging.
        """
        Consumer.log_debug("_get_message_task_callback %r", args)

    def _read_done(
        self, get_message_task: asyncio.Task, read_result: PayloadFuture
    ) -> None:
        """
        Callback for read()
        """
        Consumer.log_debug(
            "_read_done(%s) %r %r", self.consumer_name, get_message_task, read_result
        )

        # if the read_result has been canceled, cancel the task for getting
        # the next message
        if read_result.cancelled():
            Consumer.log_debug(f"   - read(%s) is canceled", self.consumer_name)
            get_message_task.cancel()
            Consumer.log_debug(f"   - %r is canceled", get_message_task)

    async def get_message(self, read_result: PayloadFuture) -> None:
        """
        Get the next message in the stream, checking the backlog first to see
        if there are any previously delivered messages that haven't been acked
        (like the consumer crashed processing the message).
        """
        Consumer.log_debug("get_message(%s) %r", self.consumer_name, read_result)

        while True:
            # if we are checking the backlog, get the next message otherwise
            # get the next one that hasn't been delivered to antoher consumer
            if self.check_backlog:
                latest_id = self.latest_id
            else:
                latest_id = b">"
            Consumer.log_debug("    - latest_id: %r", latest_id)

            args = {
                "groupname": self.group_name,
                "consumername": self.consumer_name,
                "count": 1,
                "block": self.xread_timeout,
                "streams": {self.stream_name: latest_id},
            }
            messages = None
            try:
                messages = await self.client.redis.xreadgroup(**args)
                Consumer.log_debug("    - messages: %r", messages)
            except Exception as err:
                Consumer.log_debug("    - xreadgroup exception: %r", err)

            if not messages:
                Consumer.log_debug("    - timeout")
                continue

            # no more messages means we have completely consumed the backlog
            stream, element_list = messages[0]
            if not element_list:
                self.check_backlog = False
            else:
                break

        msg_id, payload = element_list[0]
        payload_dict = dict(payload)
        Consumer.log_debug(
            "    - stream %s, id %s, payload_dict %s", stream, msg_id, payload_dict
        )

        # save the message ID so the next time this is entered it gets the
        # next message in the backlog
        self.latest_id = msg_id

        # build a Payload wrapper around the message
        payload = Payload(self, msg_id, payload_dict)
        Consumer.log_debug("    - payload: %r", payload)

        # return this payload back to the application
        read_result.set_result(payload)


@debugging
class Payload:
    """
    Encapsulates the payload wrapped around a message and exposes an ack()
    function.
    """

    consumer: Consumer
    msg_id: str
    response_channel: Optional[str]
    message: Dict[str, Any]

    log_debug: Callable[..., None]

    def __init__(
        self, consumer: Consumer, msg_id: str, payload_dict: Dict[str, Any]
    ) -> None:
        Payload.log_debug("__init__ %r %r", msg_id, payload_dict)

        self.consumer = consumer
        self.msg_id = msg_id
        self.response_channel = payload_dict.get("response_channel", None)
        try:
            self.message = json.loads(payload_dict["message"])
        except json.decoder.JSONDecodeError:
            Payload.log_debug("    - unable to decode message, log this event")
            asyncio.ensure_future(
                self.consumer.client.redis.xack(
                    self.consumer.stream_name, self.consumer.group_name, self.msg_id
                )
            )
            return

    async def ack(self, response: Any = None, error: Any = None) -> None:
        """
        Acks the message on the stream and publishes the response on the
        responseChannel, if provided.
        """
        Payload.log_debug("ack response=%r error=%r", response, error)

        Payload.log_debug("    - msg_id: %r", self.msg_id)
        await self.consumer.client.redis.xack(
            self.consumer.stream_name, self.consumer.group_name, self.msg_id
        )
        Payload.log_debug("    - xack complete")

        if self.response_channel is not None:
            Payload.log_debug("    - response channel: %r", self.response_channel)
            m_response = {"message": response, "error": error}
            await self.consumer.client.redis.publish(
                self.response_channel, json.dumps(m_response)
            )
            Payload.log_debug("    - published json")


if TYPE_CHECKING:
    # class is declared as generic in stubs but not at runtime
    PayloadFuture = asyncio.Future[Payload]  # type: ignore
else:
    PayloadFuture = asyncio.Future



"""
Producer for RedisMQ
"""
from __future__ import annotations

import asyncio
import json
import time

from typing import TYPE_CHECKING, Any, Callable, Dict, TypedDict, Optional, cast
from redis import Connection  # type: ignore[attr-defined]
from .debugging import debugging

# class is declared as generic in stubs but not at runtime
AnyFuture = asyncio.Future

# settings
MAXLEN = 100
TIMEOUT = 10.0


class Client(TypedDict):
    redis: Connection


@debugging
class Producer:
    """
    Produces messages
    """

    client: Client
    stream_name: str
    channel_key: str
    maxlen: int
    timeout: float
    id: int

    log_debug: Callable[..., None]

    def __init__(
        self,
        client: Client,
        stream_name: str,
        maxlen: int = MAXLEN,
        timeout: float = TIMEOUT,
    ) -> None:
        """
        default constructor
        """
        Producer.log_debug("__init__ %r %r", client, stream_name)

        self.client = client
        self.stream_name = stream_name
        self.channel_key = "%s:responseid" % client.namespace
        self.maxlen = maxlen
        self.timeout = timeout

        self.id = time.time_ns()

    # make the handler for the channel
    def get_handler(self, channel_id, fut: AnyFuture):
        Producer.log_debug("get_handler channel_id %r fut %r" % (channel_id, fut))

        async def _handler(json_message=None):
            Producer.log_debug("_handler json_message: %r", json_message)
            response = None
            try:
                if json_message:
                    response = json.loads(json_message["data"])
            except ValueError as err:
                Producer.log_debug("    - value/decoding error %s", channel_id)
                response = {"message": "JSON Decoding Error", "err": err}
            except TypeError as err:
                Producer.log_debug("    - type error %s", channel_id)
                response = {"message": "Type Error", "err": err}
            finally:
                Producer.log_debug("    - finally %s", channel_id)
                await self.client.pubsub.unsubscribe(channel_id)
                Producer.log_debug("    - unsubscribed %s", channel_id)
                if not fut.cancelled():
                    Producer.log_debug("    -setting result %s", response)
                    fut.set_result(response)

        return _handler

    # pylint: disable=invalid-name
    def addUnconfirmedMessage(
        self, message: Any, response_channel_id: str = None
    ) -> AnyFuture:
        """
        Return a task that adds an unconfirmed message to the message queue.
        """
        Producer.log_debug("addUnconfirmedMessage %r", message)

        # JSON encode the message
        payload = {"message": json.dumps(message)}
        if response_channel_id is not None:
            payload["response_channel"] = response_channel_id

        # create a task to add it to the stream
        future = self.client.redis.xadd(self.stream_name, payload, maxlen=self.maxlen)

        return cast(AnyFuture, future)

    # pylint: disable=invalid-name
    async def addConfirmedMessage(self, message: Any):
        """
        Adds a confirmed message to the message queue and
        results in the confirmed response.
        """
        Producer.log_debug("addConfirmedMessage %r", message)

        # JSON encode the message
        payload = {"message": json.dumps(message)}
        # create a future
        future = asyncio.get_running_loop().create_future()

        # get a unique channel identifier
        uid = await self.client.redis.incr(self.channel_key)
        response_channel_id = "%s:response.%d" % (self.client.namespace, uid)
        Producer.log_debug("    - response_channel_id: %r", response_channel_id)

        # pack it into the request payload
        payload["response_channel"] = response_channel_id

        # start listening for a response
        _handler = self.get_handler(response_channel_id, future)
        kwargs = {response_channel_id: _handler}
        try:
            await self.client.pubsub.subscribe(**kwargs)
            Producer.log_debug("    - subscribed")

            # put the request into the stream
            message_id: str = await self.client.redis.xadd(
                self.stream_name, payload, maxlen=self.maxlen
            )
            Producer.log_debug("    - message_id: %r", message_id)
            # future will get the result set by the handler when the response is published
            resp = await asyncio.wait_for(future, self.timeout)
        except asyncio.TimeoutError as err:
            Producer.log_debug("    - timeout waiting for future: %r", err)
            await _handler()
            resp = {"message": "Timeout Error", "err": err}
        except asyncio.CancelledError as err:
            Producer.log_debug("    - cancelled %r", err)
            await _handler()
            resp = {"message": "Cancelled Error", "err": err}
        except BaseException as err:
            # Producer.log_debug("    - unexpected error %r", err)
            await _handler()
            resp = {"message": "Unexpected Error", "err": err}
        return resp

    # pylint: disable=invalid-name
    def destroy(self) -> None:
        """
        Stops this producer from working. This is automatically called when
        client.dispose_producer() is called with this producer.
        """
        Producer.log_debug("destroy")

        # assume this hasn't been gracefully closed
        # close_task = asyncio.create_task(self.close())
        # Producer.log_debug("    - close task: %r", close_task)



"""
Test Debugging
"""

import pytest  # type: ignore
from redismq.debugging import debugging, create_log_handler
import logging


@debugging
class DebugTester:
    def __init__(self) -> None:
        root_logger = logging.getLogger(None)
        a_handler = logging.StreamHandler()
        DebugTester.log_debug("")
        create_log_handler('test_debugging.DebugTester',level='WARN')
        create_log_handler(handler=a_handler)
        root_logger.removeHandler(a_handler)

@pytest.mark.asyncio  # type: ignore[misc]
async def test_debugging() -> None:
    "test debugging"
    t = DebugTester()



"""
Client for RedisMQ
"""

from __future__ import annotations

import asyncio
from redis import asyncio as aioredis  # type: ignore[attr-defined]

from typing import Any, Callable, Dict, Set, Optional

from .debugging import debugging
from .producer import Producer
from .consumer import Consumer

__all__ = ["Client"]


# producer default settings
MAXLEN = 100
TIMEOUT = 10.0


@debugging
class Client:
    """
    Top level class to interface with message queue
    """

    log_debug: Callable[..., None]

    status: str
    namespace: str
    redis: Any
    pubsub: Any
    sub_task: Any

    payloads: Set[Any]
    payloads_updated: asyncio.Condition

    producer_registry: Dict[str, Producer]

    def __init__(self) -> None:
        """
        default constructor - use connect() instead
        """
        Client.log_debug("__init__")

        self.namespace = "rmq"
        self.producer_registry = {}
        self.status = "wait"

        # keep track of the un-acked payloads
        self.payloads = set()
        self.payloads_event = asyncio.Event()
        self.payloads_event.set()

    @classmethod
    async def connect(cls, address: str, namespace: Optional[str] = None) -> "Client":
        """
        Call to create a connection pool.
        """
        Client.log_debug("connect %s", address)

        # create a Client instance or one of its subclasses
        client = cls()

        # set the namespace
        if namespace:
            cls.namespace = namespace

        # we are 'connecting' but not really until the PING
        client.status = "connecting"

        # create a blocking connection pool to wait for a connection to become available
        # rather than raising an exception
        # see https://aioredis.readthedocs.io/en/latest/api/low-level/#aioredis.connection.BlockingConnectionPool
        pool = aioredis.BlockingConnectionPool.from_url(
            address, max_connections=10, decode_responses=True
        )
        client.redis = aioredis.Redis(connection_pool=pool)
        Client.log_debug("    - redis: %s", client.redis)

        # try to ping it
        rslt = await client.redis.ping()
        Client.log_debug("    - ping: %r", rslt)

        client.pubsub = client.redis.pubsub(ignore_subscribe_messages=True)
        client.sub_task = asyncio.get_running_loop().create_task(client.pubsub.run())
        client.status = "ready"

        return client

    async def close(self) -> None:
        """
        Call to wait for all of the active payloads to complete.
        """
        Client.log_debug("close")
        if self.status in ["closed", "closing"]:
            raise RuntimeError("Client.close() has already been called")
        if self.status != "ready":
            raise RuntimeError("Client is not ready to close")

        self.status = "closing"

        # wait for the event that says no more pending
        Client.log_debug(f"    - payloads: {self.payloads}")
        await self.payloads_event.wait()
        self.sub_task.cancel()
        try:
            await self.sub_task
        except asyncio.CancelledError:
            Client.log_debug(f"    - sub_task cancelled")
        await self.pubsub.aclose()
        Client.log_debug(f"    - pubsub closed")
        await self.redis.aclose()
        Client.log_debug(f"    - redis closed")
        await self.redis.connection_pool.disconnect()
        Client.log_debug(f"    - connection_pool disconnected")

        self.status = "closed"

    async def producer(
        self, stream_name: str, maxlen: int = MAXLEN, timeout: float = TIMEOUT
    ) -> Producer:
        """
        Use this to get a Producer
        """
        Client.log_debug("producer %s", stream_name)
        if stream_name not in self.producer_registry:
            Client.log_debug("    - adding producer %s to registry", stream_name)
            self.producer_registry[stream_name] = Producer(
                self, stream_name, maxlen, timeout
            )
        else:
            Client.log_debug("    - producer %s found in registry", stream_name)
        return self.producer_registry[stream_name]

    async def dispose_producer(self, producer: Producer) -> None:
        """dispose_producer of client
        @param producer - the producer to dispose of
        """
        stream_name = producer.stream_name
        if self.producer_registry[stream_name] is producer:
            del self.producer_registry[stream_name]
            producer.destroy()
            Client.log_debug("dispose_producer - producer %s disposed", stream_name)
        else:
            raise ValueError("Producer named %s registry entry mismatch." % stream_name)

    async def consumer(
        self,
        stream_name: str,
        group_name: str,
        consumer_id: str,
        scan_pending_on_start: bool = True,
        claim_stale_messages: bool = True,
        min_idle_time: int = 60000,
    ) -> Consumer:
        """
        Use this to get a Consumer
        """
        Client.log_debug("consumer %s ...", stream_name)

        try:
            group_info = await self.redis.xinfo_groups(stream_name)
            Client.log_debug("    - xinfo_groups: %r", group_info)
        except aioredis.RedisError as err:
            Client.log_debug("    - no existing group: %r", err)
            group_info = []

        if not any(e["name"] == group_name for e in group_info):
            await self.redis.xgroup_create(
                stream_name, group_name, id="$", mkstream=True
            )
            Client.log_debug(
                "    - added group %s to stream %s ", group_name, stream_name
            )

        # create a consumer
        consumer = Consumer(
            self,
            stream_name,
            group_name,
            consumer_id,
            min_idle_time,
        )

        try:
            stream_info = await self.redis.xinfo_stream(stream_name)
            Client.log_debug("    - xinfo_stream: %r", stream_info)

            # set the consumer to read new messages that haven't
            # been delivered to another consumer
            consumer.latest_id = b">"

        except aioredis.RedisError as err:
            Client.log_debug("    - no existing stream: %r", err)
            stream_info = []

        if scan_pending_on_start:
            rslt = await self.redis.xpending(stream_name, group_name)
            Client.log_debug(f"    - xpending: %r", rslt)

            pending_count = rslt["pending"]
            min_id = rslt["min"]
            max_id = rslt["max"]
            pending_consumers = rslt["consumers"]
            if not pending_count:
                Client.log_debug("    - no pending messages")

            for pending_info in pending_consumers:
                pending_consumer = pending_info["name"]
                pending_consumer_count = pending_info["pending"]
                if pending_consumer == consumer_id:
                    Client.log_debug(
                        f"    - this consumer has {pending_consumer_count} pending messages"
                    )
                else:
                    Client.log_debug(f"    - pending messages for {pending_consumer!r}")

                pending_messages = await self.redis.xpending_range(
                    stream_name,
                    group_name,
                    min=b"-",
                    max=b"+",
                    count=pending_consumer_count,
                    consumername=pending_consumer,
                )
                Client.log_debug(f"    - pending_messages: %r", pending_messages)

                for pending_message_info in pending_messages:
                    message_id = pending_message_info["message_id"]
                    pending_consumer = pending_message_info["consumer"]
                    time_since_delivered = pending_message_info["time_since_delivered"]
                    times_delivered = pending_message_info["times_delivered"]

                    Client.log_debug(
                        "        %r for %r, idle %rs, delivered %r times",
                        message_id,
                        pending_consumer,
                        time_since_delivered / 1000.0,
                        times_delivered,
                    )
                    if claim_stale_messages:
                        retcode = await self.redis.xclaim(
                            stream_name,
                            group_name,
                            consumer_id,
                            min_idle_time,
                            [message_id],
                        )
                        Client.log_debug(f"        claim: {retcode!r}")

                        # start at the beginning of the pending messages
                        consumer.latest_id = b"0-0"
                        consumer.check_backlog = True

        return consumer



"""
RedisMQ
"""

import os
import logging

__version__ = "1.1.8"

_log = logging.getLogger(__name__)

from .debugging import create_log_handler

create_log_handler(_log, level=os.getenv("REDISMQ", logging.WARNING))

from .client import Client
from .producer import Producer
from .consumer import Consumer

__all__ = ["Client", "Producer", "Consumer"]



"""
Test Client
"""

import pytest  # type: ignore
from redis import exceptions # type: ignore[attr-defined]

from tests.utils import TEST_URL  # type: ignore
from redismq import Client


@pytest.mark.asyncio  # type: ignore[misc]
async def test_client_connect() -> None:
    "test client with valid options"
    p_connection = await Client.connect(TEST_URL, namespace='foo')
    await p_connection.close()


@pytest.mark.asyncio  # type: ignore[misc]
async def test_client_barf() -> None:
    "test client with invalid URL should throw"
    with pytest.raises(exceptions.ConnectionError):
        await Client.connect("redis://bogus")


@pytest.mark.asyncio  # type: ignore[misc]
async def test_multiple_client_close() -> None:
    "multiple clients should open/close gracefully"
    p_connection = await Client.connect(TEST_URL)
    q_connection = await Client.connect(TEST_URL)
    await q_connection.close()
    await p_connection.close()


@pytest.mark.asyncio  # type: ignore[misc]
async def test_client_close_twice() -> None:
    "client closed twice should throw"
    p_connection = await Client.connect(TEST_URL)
    await p_connection.close()
    with pytest.raises(RuntimeError):
        await p_connection.close()



"""
Test Unconfirmed Messages
"""
import pytest  # type: ignore

from redismq import Client
from tests.utils import TEST_URL  # type: ignore


@pytest.mark.asyncio  # type: ignore[misc]
async def test_send_an_unconfirmed_message() -> None:
    "test unconfirmed message"
    mq_connection = await Client.connect(TEST_URL)
    my_producer = await mq_connection.producer("mystream")
    response = await my_producer.addUnconfirmedMessage("Hello there!")
    print("Got message ID", response)
    await mq_connection.close()


@pytest.mark.asyncio  # type: ignore[misc]
async def test_same_producer_twice() -> None:
    "test same producer"
    mq_connection = await Client.connect(TEST_URL)
    my_producer = await mq_connection.producer("mystream")
    your_producer = await mq_connection.producer("mystream")
    assert my_producer is your_producer
    await mq_connection.close()



"""
Test Confirmed Messages
"""
import asyncio
import pytest  # type: ignore
from tests.utils import TEST_URL  # type: ignore
from redismq.debugging import debugging
from redismq import Client, Producer, Consumer


async def send_a_confirmed_message(my_producer: Producer, delay: int = 0) -> None:
    "test confirmed message"
    await asyncio.sleep(delay)
    response = await my_producer.addConfirmedMessage(
        "Hello there! Let me know when you get this."
    )
    assert response["message"] == "I got your message"


async def read_a_confirmed_message(my_consumer: Consumer) -> None:
    "test reading a confirmed message"
    payload = await my_consumer.read()
    resp = "I got your message" if payload.response_channel else "no response"
    await payload.ack(resp)


async def botch_a_confirmed_message(my_conn, my_stream, my_group, my_consumer) -> None:
    "return bogus response to a confirmed message"

    args = {
        "groupname": my_group,
        "consumername": my_consumer,
        "count": 1,
        "block": 1000,
        "streams": {my_stream: b">"},
    }
    while True:
        messages = await my_conn.xreadgroup(**args)
        if not messages:
            continue
        stream, element_list = messages[0]
        msg_id, payload = element_list[0]
        payload_dict = dict(payload)
        response_channel = payload_dict.get("response_channel", None)
        # ack the message in the stream
        await my_conn.xack(my_stream, my_group, msg_id)
        if response_channel is not None:
            await my_conn.publish(response_channel, "not json")
            return


async def receive_a_botched_confirmation(my_producer: Producer, delay: int = 0) -> None:
    "test botched confirmation"
    await asyncio.sleep(delay)
    response = await my_producer.addConfirmedMessage(
        "Hello there! Let me know when you get this."
    )
    assert response["message"] == "JSON Decoding Error"
    assert "err" in response


async def ack_confirmed_messages(my_consumer: Consumer) -> None:
    "ack confirmed messages"
    while True:
        try:
            payload = await my_consumer.read()
            await payload.ack(f"Acknowledged {payload.message}")
        except Exception as err:
            raise


@pytest.mark.asyncio  # type: ignore[misc]
async def test_send_and_read() -> None:
    "test sending a confirmed message and reading/confirming it"
    p_connection = await Client.connect(TEST_URL)
    await p_connection.redis.delete("mystream")
    my_producer = await p_connection.producer("mystream")
    # add a message so pending messages will get processed
    await my_producer.addUnconfirmedMessage("Hello there!")
    q_connection = await Client.connect(TEST_URL)
    my_consumer = await q_connection.consumer("mystream", "mygroup", "consumer1")

    send_task = asyncio.create_task(send_a_confirmed_message(my_producer))
    recv_task = asyncio.create_task(read_a_confirmed_message(my_consumer))
    await asyncio.gather(send_task, recv_task)
    recv_task.cancel()
    await p_connection.dispose_producer(my_producer)

    await p_connection.close()
    await q_connection.close()


@pytest.mark.asyncio  # type: ignore[misc]
async def test_pending() -> None:
    "test processing pending messages"
    p_connection = await Client.connect(TEST_URL)
    q_connection = await Client.connect(TEST_URL)
    await p_connection.redis.delete("mystream")
    my_producer = await p_connection.producer("mystream")
    await q_connection.consumer("mystream", "mygroup", "consumer0")
    # add a message
    await my_producer.addUnconfirmedMessage("Hello there!")
    # xread the message so it looks like it was read by consumer1 but never processed
    await p_connection.redis.xreadgroup(
        groupname="mygroup",
        consumername="consumer1",
        count=1,
        block=500,
        streams={"mystream": b">"},
    )
    await q_connection.consumer("mystream", "mygroup", "consumer1")
    await p_connection.dispose_producer(my_producer)

    await p_connection.close()
    await q_connection.close()


@pytest.mark.execution_timeout(10)
@pytest.mark.asyncio  # type: ignore[misc]
async def test_slow_send_and_read() -> None:
    "test sending a slow confirmed message and reading/confirming it"
    p_connection = await Client.connect(TEST_URL)
    await p_connection.redis.delete("mystream")
    my_producer = await p_connection.producer("mystream")
    q_connection = await Client.connect(TEST_URL)
    my_consumer = await q_connection.consumer("mystream", "mygroup", "consumer1")

    send_task = asyncio.create_task(send_a_confirmed_message(my_producer, 5))
    recv_task = asyncio.create_task(read_a_confirmed_message(my_consumer))
    await asyncio.gather(send_task, recv_task)
    recv_task.cancel()

    await p_connection.close()
    await q_connection.close()


@pytest.mark.asyncio  # type: ignore[misc]
async def test_multiple_confirmed() -> None:
    "test many confirmed messages"
    p_connection = await Client.connect(TEST_URL)
    await p_connection.redis.delete("mystream")
    my_producer = await p_connection.producer("mystream")

    q_connection = await Client.connect(TEST_URL)
    my_consumer = await q_connection.consumer("mystream", "mygroup", "consumer1")

    ack_task = asyncio.create_task(ack_confirmed_messages(my_consumer))
    for i in range(10):
        payload = f"message {i}"
        response = await my_producer.addConfirmedMessage(payload)
        assert response["message"] == f"Acknowledged {payload}"
    ack_task.cancel()

    await p_connection.close()
    await q_connection.close()


@pytest.mark.asyncio  # type: ignore[misc]
async def test_concurrent_confirmed() -> None:
    "test many concurrent confirmed messages"
    p_connection = await Client.connect(TEST_URL)
    await p_connection.redis.delete("mystream")
    my_producer = await p_connection.producer("mystream")

    q_connection = await Client.connect(TEST_URL)
    my_consumer = await q_connection.consumer("mystream", "mygroup", "consumer1")
    confirmed_coros = []
    for i in range(20):
        # response = await my_producer.addConfirmedMessage(f"message {i}")
        confirmed_coros.append(my_producer.addConfirmedMessage(f"message {i}"))
    ack_task = asyncio.create_task(ack_confirmed_messages(my_consumer))
    responses = await asyncio.gather(*confirmed_coros)
    assert all(
        [responses[i]["message"] == f"Acknowledged message {i}" for i in range(20)]
    )
    ack_task.cancel()

    await p_connection.close()
    await q_connection.close()


@pytest.mark.asyncio  # type: ignore[misc]
async def test_bad_json() -> None:
    "test sending a confirmed message receiving a bogus response"
    p_connection = await Client.connect(TEST_URL)
    await p_connection.redis.delete("mystream")
    my_producer = await p_connection.producer("mystream")
    q_connection = await Client.connect(TEST_URL)
    # call client.consumer just to create the consumer group
    await q_connection.consumer("mystream", "mygroup", "consumer1")

    # add a message
    await my_producer.addUnconfirmedMessage("Hello there!")

    send_task = asyncio.create_task(receive_a_botched_confirmation(my_producer))
    recv_task = asyncio.create_task(
        botch_a_confirmed_message(
            q_connection.redis, "mystream", "mygroup", "consumer1"
        )
    )
    await asyncio.gather(send_task, recv_task)
    recv_task.cancel()
    await p_connection.dispose_producer(my_producer)

    await p_connection.close()
    await q_connection.close()


@pytest.mark.asyncio  # type: ignore[misc]
async def test_confirmed_timeout() -> None:
    "test timeout confirmed message"
    p_connection = await Client.connect(TEST_URL)
    await p_connection.redis.delete("mystream")
    my_producer = await p_connection.producer("mystream", timeout=0.1)
    # no consumer
    resp = await my_producer.addConfirmedMessage(f"message to nowhere")
    assert resp["message"] == "Timeout Error"
    assert "err" in resp
    # commented out because redis does not reliably report no subscribers
    # even though the producer unsubscribed
    #    channels = await p_connection.redis.pubsub_channels()
    #    for channel in channels:
    # to force redis to update the channel list for this connection
    #        await p_connection.redis.publish(channel, "blah")
    #        assert (await p_connection.redis.execute_command('PUBSUB', 'NUMSUB', channel))[1] == 0
    await p_connection.close()


@debugging
@pytest.mark.asyncio  # type: ignore[misc]
async def test_cancelled_confirmed() -> None:
    "test cancelling a confirmed message"
    p_connection = await Client.connect(TEST_URL)
    await p_connection.redis.delete("mystream")
    my_producer = await p_connection.producer("mystream")
    # no consumer
    coro = my_producer.addConfirmedMessage("msg")
    try:
        await asyncio.wait_for(coro, timeout=0.01)
    except asyncio.TimeoutError:
        pass
    # commented out because redis does not reliably report no subscribers
    # even though the producer unsubscribed
    #        channels = await p_connection.redis.pubsub_channels()
    #        while channels:
    #            print(f"addConfirmed running: {coro.cr_running} channels: {channels}")
    #            await asyncio.sleep(.01)
    #            channels = await p_connection.redis.pubsub_channels()
    await p_connection.close()
